/* This file is generated by glib-genmarshal, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */
#ifndef ___MOUSEPAD_MARSHAL_MARSHAL_H__
#define ___MOUSEPAD_MARSHAL_MARSHAL_H__

#include <glib-object.h>

G_BEGIN_DECLS

/* VOID:INT,INT,INT (mousepad-marshal.list:1) */
extern
void _mousepad_marshal_VOID__INT_INT_INT (GClosure     *closure,
                                          GValue       *return_value,
                                          guint         n_param_values,
                                          const GValue *param_values,
                                          gpointer      invocation_hint,
                                          gpointer      marshal_data);

/* VOID:INT,INT,STRING,FLAGS (mousepad-marshal.list:2) */
extern
void _mousepad_marshal_VOID__INT_INT_STRING_FLAGS (GClosure     *closure,
                                                   GValue       *return_value,
                                                   guint         n_param_values,
                                                   const GValue *param_values,
                                                   gpointer      invocation_hint,
                                                   gpointer      marshal_data);

/* VOID:FLAGS,STRING,STRING (mousepad-marshal.list:3) */
extern
void _mousepad_marshal_VOID__FLAGS_STRING_STRING (GClosure     *closure,
                                                  GValue       *return_value,
                                                  guint         n_param_values,
                                                  const GValue *param_values,
                                                  gpointer      invocation_hint,
                                                  gpointer      marshal_data);

/* VOID:OBJECT,INT,INT (mousepad-marshal.list:4) */
extern
void _mousepad_marshal_VOID__OBJECT_INT_INT (GClosure     *closure,
                                             GValue       *return_value,
                                             guint         n_param_values,
                                             const GValue *param_values,
                                             gpointer      invocation_hint,
                                             gpointer      marshal_data);


G_END_DECLS

#endif /* ___MOUSEPAD_MARSHAL_MARSHAL_H__ */
